package com.sun.grizzly;

import junit.framework.TestCase;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.net.Socket;
import java.net.BindException;
import java.net.SocketAddress;
import java.io.IOException;
import java.util.logging.Level;

import com.sun.grizzly.utils.ControllerUtils;

/**
 * @author Bongjae Chang
 * @date 2009. 6. 15
 */
public class RecycledChannelTest extends TestCase {

    private static final int PORT = 17520;
    private static final int SLEEP_TIME = 3000; // ms

    private final InetAddress localInetAddress;
    private final InetSocketAddress localInetSocketAddress;

    public RecycledChannelTest() throws UnknownHostException {
        localInetAddress = InetAddress.getLocalHost();
        localInetSocketAddress = new InetSocketAddress( localInetAddress, PORT );
    }

    public void testSimpleTCPConnect() throws IOException {
        final Controller controller = new Controller();
        SelectorHandler selectorHandler = new TCPSelectorHandler();
        ( (TCPSelectorHandler)selectorHandler ).setPort( PORT );
        ( (TCPSelectorHandler)selectorHandler ).setInet( localInetAddress );
        ( (TCPSelectorHandler)selectorHandler ).setReuseAddress( false );
        controller.addSelectorHandler( selectorHandler );

        Socket clientSocket = null;
        try {
            ControllerUtils.startController( controller );

            boolean result = false;
            Controller.logger().log( Level.INFO, "Try to get a connector handler with the local address which already has been bound." );
            try {
                tryToConnect( controller, Controller.Protocol.TCP, null, localInetSocketAddress );
            } catch( IOException ie ) {
                if( ie instanceof BindException ) {
                    result = true;
                    Controller.logger().log( Level.INFO, "Got the expected BindException." );
                    assertTrue( "Got the expected BindException.", true );
                } else {
                    Controller.logger().log( Level.INFO, "Got the unexpected error.", ie );
                    assertTrue( "Got the unexpected error.", false );
                }
            }
            if( !result )
                assertTrue( "The BindException was expected.", false );

            Controller.logger().log( Level.INFO, "Try to connect the local server." );
            clientSocket = new Socket( localInetAddress, PORT );
            Controller.logger().log( Level.INFO, "Wait for " + SLEEP_TIME + "(ms)" );
            try {
                Thread.sleep( SLEEP_TIME );
            } catch( InterruptedException e ) {
            }

            Controller.logger().log( Level.INFO, "Try to get a connector handler with the local address which already has been bound again." );
            try {
                tryToConnect( controller, Controller.Protocol.TCP, clientSocket.getLocalSocketAddress(), localInetSocketAddress );
            } catch( IOException ie ) {
                Controller.logger().log( Level.INFO, "Got the unexpected error.", ie );
                assertTrue( "Got the unexpected error.", false );
                throw ie;
            }
        } finally {
            if( clientSocket != null ) {
                try {
                    clientSocket.shutdownInput();
                } catch( IOException e ) {
                }
                try {
                    clientSocket.shutdownOutput();
                } catch( IOException e ) {
                }
                try {
                    clientSocket.close();
                } catch( IOException e ) {
                }
            }
            controller.stop();
        }
    }

    public void testSimpleUDPConnect() throws IOException {
        final Controller controller = new Controller();
        SelectorHandler selectorHandler = new UDPSelectorHandler();
        ( (UDPSelectorHandler)selectorHandler ).setPort( PORT );
        ( (UDPSelectorHandler)selectorHandler ).setInet( localInetAddress );
        ( (UDPSelectorHandler)selectorHandler ).setReuseAddress( false );
        controller.addSelectorHandler( selectorHandler );

        try {
            ControllerUtils.startController( controller );

            Controller.logger().log( Level.INFO, "Try to get a connector handler with the local address which already has been bound." );
            try {
                tryToConnect( controller, Controller.Protocol.UDP, localInetSocketAddress, localInetSocketAddress );
            } catch( IOException ie ) {
                Controller.logger().log( Level.INFO, "Got the unexpected error.", ie );
                assertTrue( "Got the unexpected error.", false );
                throw ie;
            }
        } finally {
            controller.stop();
        }
    }

    private void tryToConnect( Controller controller, Controller.Protocol protocol, SocketAddress remote, SocketAddress local ) throws IOException {
        ConnectorHandler connectorHandler = null;
        try {
            connectorHandler = controller.acquireConnectorHandler( protocol );
            connectorHandler.connect( remote, local );
        } finally {
            if( connectorHandler != null ) {
                try {
                    connectorHandler.close();
                } catch( IOException e ) {
                    e.printStackTrace();
                }
                controller.releaseConnectorHandler( connectorHandler );
            }
        }
    }

    public static void main( String[] args ) throws IOException {
        RecycledChannelTest test = new RecycledChannelTest();
        test.testSimpleTCPConnect();
        test.testSimpleUDPConnect();
    }
}
